
from __future__ import annotations

# packages
import numpy as np
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from numba import njit
from dataclasses import dataclass
from typing import Union, Dict, Any, Optional, Tuple, List, NamedTuple
from enum import Enum

# qis
import qis.file_utils as fu
import qis.plots.derived.regime_data
import qis.utils.dates as da
import qis.utils.df_groups as dfg
import qis.utils.struct_ops as sop
from qis.utils import EnumMap
from qis.perfstats.config import PerfParams, RegimeData
import qis.perfstats.returns as ret
import qis.perfstats.perf_stats as rpt
import qis.perfstats.regime_classifier as rcl
from qis.perfstats.regime_classifier import BenchmarkReturnsQuantileRegimeSpecs

import qis.models.linear.ewm_factors as ef
import qis.plots.time_series as pts
import qis.plots.stackplot as pst
import qis.plots.derived.prices as ppd
import qis.plots.derived.perf_table as ppt
import qis.plots.derived.returns_scatter as prs
import qis.plots.derived.returns_heatmap as rhe
from qis.plots.derived.returns_heatmap import plot_returns_heatmap


class MetricSpec(NamedTuple):
    title: str


class AttributionMetric(MetricSpec, Enum):
    PNL = MetricSpec(title='P&L Attribution, sum=portfolio performance')
    PNL_RISK = MetricSpec(title='P&L Risk Attribution, sum=100%')
    INST_PNL = MetricSpec(title='Instrument P&L')


@dataclass
class PortfolioData:
    nav: pd.Series  # nav is computed with all cost
    weights: pd.DataFrame = None  # weights of portfolio
    units: pd.DataFrame = None  # units of portfolio instruments
    prices: pd.DataFrame = None  # prices of portfolio universe
    instrument_pnl: pd.DataFrame = None  # include net pnl by intsrument
    realized_costs: pd.DataFrame = None  # realized trading costs by instrument
    input_weights: Union[np.ndarray, pd.DataFrame, Dict[str, float]] = None  # inputs to potfolio
    is_rebalancing: pd.Series = None  # optional rebal info
    tickers_to_names_map: Optional[Dict[str, str]] = None  # renaming of long tickers
    group_data: pd.Series = None  # for asset class grouping
    group_order: List[str] = None

    def __post_init__(self):

        if isinstance(self.nav, pd.DataFrame):
            self.nav = self.nav.iloc[:, 0]
        if self.prices is None:
            self.prices = self.nav.to_frame()
        if self.weights is None:  # default will be delta-1 portfolio of nav
            self.weights = pd.DataFrame(1.0, index=self.prices.index, columns=self.prices.columns)
        if self.units is None:  # default will be delta-1 portfolio of nav
            self.units = pd.DataFrame(1.0, index=self.prices.index, columns=self.prices.columns)
        if self.instrument_pnl is None:
            self.instrument_pnl = self.prices.pct_change().multiply(self.weights.shift(1)).fillna(0.0)
        if self.realized_costs is None:
            self.realized_costs = pd.DataFrame(0.0, index=self.prices.index, columns=self.prices.columns)
        if self.group_data is None:  # use instruments as groups
            self.group_data = pd.Series(self.prices.columns, index=self.prices.columns)
        if self.group_order is None:
            self.group_order = list(self.group_data.unique())

    def save(self, ticker: str) -> None:
        datasets = dict(nav=self.nav, prices=self.prices, weights=self.weights, units=self.units,
                        instrument_pnl=self.instrument_pnl, realized_costs=self.realized_costs)
        if self.group_data is not None:
            datasets['group_data'] = self.group_data
        fu.save_df_dict_to_csv(datasets=datasets, file_name=ticker)
        print(f"saved portfolio data for {ticker}")

    @classmethod
    def load(cls, ticker: str) -> PortfolioData:
        dataset_keys = ['nav', 'prices', 'weights', 'units', 'instrument_pnl', 'realized_costs', 'group_data']
        datasets = fu.load_df_dict_from_csv(dataset_keys=dataset_keys, file_name=ticker)
        return cls(**datasets)

    def _set_group_data(self, group_data: pd.Series, group_order: List[str] = None) -> None:
        self.group_data = group_data
        self.group_order = group_order

    """
    NAV level getters
    """
    def get_portfolio_nav(self, time_period: da.TimePeriod = None) -> pd.Series:
        """
        get nav using consistent function for all return computations
        """
        if time_period is not None:
            nav_ = time_period.locate(self.nav)
        else:
            nav_ = self.nav
        return nav_

    def get_instruments_pnl(self,
                            time_period: da.TimePeriod = None,
                            is_compounded: bool = False
                            ) -> pd.DataFrame:

        pnl = self.instrument_pnl
        if is_compounded:
            pnl = np.expm1(pnl)
        if time_period is not None:
            pnl = time_period.locate(pnl)
        return pnl

    def get_instruments_navs(self,
                             time_period: da.TimePeriod = None,
                             constant_trade_level: bool = False
                             ) -> pd.DataFrame:
        pnl = self.get_instruments_pnl(time_period=time_period)
        navs = ret.returns_to_nav(returns=pnl, constant_trade_level=constant_trade_level)
        return navs

    def get_ac_navs(self,
                    time_period: da.TimePeriod = None,
                    constant_trade_level: bool = False
                    ) -> pd.DataFrame:
        grouped_ac_pnl = dfg.agg_df_by_groups_ax1(df=self.get_instruments_pnl(time_period=time_period),
                                                  group_data=self.group_data,
                                                  agg_func=np.sum,
                                                  total_column=str(self.nav.name),
                                                  group_order=self.group_order)
        ac_navs = ret.returns_to_nav(returns=grouped_ac_pnl, constant_trade_level=constant_trade_level)
        return ac_navs

    def get_exposures(self,
                      time_period: da.TimePeriod = None,
                      is_grouped: bool = False,
                      add_total: bool = True
                      ) -> pd.DataFrame:
        if is_grouped:
            exposures = dfg.agg_df_by_groups_ax1(df=self.weights,
                                                 group_data=self.group_data,
                                                 agg_func=np.nansum,
                                                 total_column=str(self.nav.name) if add_total else None,
                                                 group_order=self.group_order)
        else:
            exposures = self.weights
        if time_period is not None:
            exposures = time_period.locate(exposures)
        return exposures

    def get_turnover(self,
                     is_agg: bool = False,
                     is_grouped: bool = False,
                     time_period: da.TimePeriod = None,
                     roll_period: Optional[int] = 260,
                     add_total: bool = True,
                     freq: Optional[str] = None
                     ) -> Union[pd.DataFrame, pd.Series]:
        turnover = (self.units.diff(1).abs()).multiply(self.prices)
        abs_exposure = self.units.multiply(self.prices).abs().sum(axis=1)
        # turnover = turnover.divide(self.nav.to_numpy(), axis=0)
        turnover = turnover.divide(abs_exposure.to_numpy(), axis=0)
        if is_agg:
            turnover = pd.Series(np.nansum(turnover, axis=1), index=self.nav.index, name=self.nav.name)
        elif is_grouped or len(turnover.columns) > 10:  # agg by groups
            turnover = dfg.agg_df_by_groups_ax1(df=turnover,
                                                group_data=self.group_data,
                                                agg_func=np.nansum,
                                                total_column=str(self.nav.name) if add_total else None,
                                                group_order=self.group_order)
        else:
            if add_total:
                turnover = pd.concat([turnover.sum(axis=1).rename(self.nav.name), turnover], axis=1)

        if roll_period is not None:
            turnover = turnover.rolling(roll_period).sum()
        elif freq is not None:
            turnover = turnover.resample(freq).sum()
        if time_period is not None:
            turnover = time_period.locate(turnover)
        return turnover

    def get_costs(self,
                  is_agg: bool = False,
                  is_grouped: bool = False,
                  time_period: da.TimePeriod = None,
                  add_total: bool = True,
                  roll_period: Optional[int] = 260
                  ) -> Union[pd.DataFrame, pd.Series]:

        costs = self.realized_costs.divide(self.nav.to_numpy(), axis=0)
        if is_agg:
            costs = pd.Series(np.nansum(costs, axis=1), index=self.nav.index, name=self.nav.name)
        elif is_grouped:
            costs = dfg.agg_df_by_groups_ax1(costs,
                                             group_data=self.group_data,
                                             agg_func=np.nansum,
                                             total_column=str(self.nav.name) if add_total else None,
                                             group_order=self.group_order)
        else:
            if add_total:
                costs = pd.concat([costs.sum(axis=1).rename(self.nav.name), costs], axis=1)

        if roll_period is not None:
            costs = costs.rolling(roll_period).sum()
        if time_period is not None:
            costs = time_period.locate(costs)
        return costs

    def compute_mcap_participation(self,
                                   mcap: pd.DataFrame,
                                   trade_level: float = 100000000
                                   ) -> pd.DataFrame:
        exposure = (self.units.multiply(self.prices)).divide(self.nav.to_numpy(), axis=0)
        participation = trade_level*exposure.divide(mcap)
        return participation

    def compute_volume_participation(self,
                                     volumes: pd.DataFrame,
                                     trade_level: float = 100000000
                                     ) -> pd.DataFrame:
        turnover = self.get_turnover(is_agg=False)
        participation = trade_level*turnover.divide(volumes)
        return participation

    def compute_cumulative_attribution(self) -> pd.DataFrame:
        attribution = (self.prices.pct_change()).multiply(self.weights.shift(1))
        attribution = attribution.cumsum(axis=0)
        return attribution

    def compute_realized_pnl(self, time_period: da.TimePeriod = None) -> Tuple[pd.DataFrame, ...]:
        avg_costs, realized_pnl, mtm_pnl, trades = compute_realized_pnl(prices=self.prices.to_numpy(),
                                                                        units=self.units.to_numpy())
        avg_costs = pd.DataFrame(avg_costs, index=self.prices.index, columns=self.prices.columns)
        realized_pnl = pd.DataFrame(realized_pnl, index=self.prices.index, columns=self.prices.columns)
        mtm_pnl = pd.DataFrame(mtm_pnl, index=self.prices.index, columns=self.prices.columns)
        trades = pd.DataFrame(trades, index=self.prices.index, columns=self.prices.columns)
        if time_period is not None:
            avg_costs = time_period.locate(avg_costs)
            realized_pnl = time_period.locate(realized_pnl)
            mtm_pnl = time_period.locate(mtm_pnl)
            trades = time_period.locate(trades)
        realized_pnl = realized_pnl.cumsum(axis=0)
        total_pnl = realized_pnl.add(mtm_pnl)
        return avg_costs, realized_pnl, mtm_pnl, total_pnl, trades

    def compute_portfolio_benchmark_betas(self, benchmark_prices: pd.DataFrame,
                                          time_period: da.TimePeriod = None,
                                          freq: str = None,
                                          span: int = 65 # quarter
                                          ) -> pd.DataFrame:
        instrument_prices = self.prices
        benchmark_prices = benchmark_prices.reindex(index=instrument_prices.index, method='ffill')
        ewm_linear_model = ef.estimate_ewm_linear_model(x=ret.to_returns(prices=benchmark_prices, freq=freq),
                                                        y=ret.to_returns(prices=instrument_prices, freq=freq),
                                                        span=span,
                                                        is_x_correlated=True)
        exposures = self.get_exposures().reindex(index=instrument_prices.index, method='ffill')
        benchmark_betas = ewm_linear_model.compute_agg_factor_exposures(asset_exposures=exposures)
        benchmark_betas = benchmark_betas.replace({0.0: np.nan}).fillna(method='ffill')  # fillholidays
        if time_period is not None:
            benchmark_betas = time_period.locate(benchmark_betas)
        return benchmark_betas

    def compute_portfolio_benchmark_attribution(self, benchmark_prices: pd.DataFrame,
                                                time_period: da.TimePeriod = None,
                                                freq: str = 'B',
                                                span: int = 63 # quarter
                                                ) -> pd.DataFrame:
        portfolio_benchmark_betas = self.compute_portfolio_benchmark_betas(benchmark_prices=benchmark_prices,
                                                                           freq=freq, span=span)
        benchmark_prices = benchmark_prices.reindex(index=portfolio_benchmark_betas.index, method='ffill')
        x = ret.to_returns(prices=benchmark_prices, freq=freq)
        x_attribution = (portfolio_benchmark_betas.shift(1)).multiply(x)
        total_attrib = x_attribution.sum(1)
        total = self.get_portfolio_nav().reindex(index=total_attrib.index, method='ffill').pct_change()
        residual = np.subtract(total, total_attrib)
        # joint_attrib = pd.concat([x_attribution, total_attrib.rename('Total benchmarks'), residual.rename('Residual')], axis=1)
        joint_attrib = pd.concat([x_attribution, residual.rename('Residual')], axis=1)
        if time_period is not None:
            joint_attrib = time_period.locate(joint_attrib)
        joint_attrib = joint_attrib.cumsum(axis=0)
        return joint_attrib

    """
    ### instrument level getters
    """
    def get_instruments_returns(self,
                                time_period: da.TimePeriod = None
                                ) -> pd.DataFrame:
        returns = self.prices.pct_change()
        if time_period is not None:
            returns = time_period.locate(returns)
        return returns

    def get_instruments_periodic_returns(self,
                                         time_period: da.TimePeriod = None,
                                         freq: str = 'M'
                                         ) -> pd.DataFrame:
        returns = self.get_instruments_returns(time_period=time_period)
        prices = ret.returns_to_nav(returns=returns, init_period=None)
        returns_f = ret.to_returns(prices=prices, freq=freq)
        return returns_f

    def get_instruments_performance_attribution(self,
                                                time_period: da.TimePeriod = None,
                                                constant_trade_level: bool = False
                                                ) -> pd.Series:
        navs = self.get_instruments_navs(time_period=time_period, constant_trade_level=constant_trade_level)
        perf = ret.to_total_returns(prices=navs).rename(self.nav.name)
        return perf

    def get_instruments_pnl_risk_attribution(self,
                                             time_period: da.TimePeriod = None
                                             ) -> pd.DataFrame:
        pnl = self.get_instruments_pnl(time_period=time_period)
        portfolio_pnl = pnl.sum(axis=1)

        pnl_risk = np.nanstd(pnl.replace({0.0: np.nan}), axis=0)
        portfolio_pnl_risk = np.nanstd(portfolio_pnl.replace({0.0: np.nan}), axis=0)
        # pnl_risk_ratio = pnl_risk / portfolio_pnl_risk
        pnl_risk_ratio = pnl_risk / np.nansum(pnl_risk)

        data = pd.Series(pnl_risk_ratio, index=pnl.columns, name=self.nav.name)
        if self.tickers_to_names_map is not None:
            data = data.rename(index=self.tickers_to_names_map)
        return data

    def get_performance_data(self,
                             attribution_metric: AttributionMetric = AttributionMetric.PNL,
                             time_period: da.TimePeriod = None
                             ) -> Union[pd.DataFrame, pd.Series]:
        if attribution_metric == AttributionMetric.PNL:
            data = self.get_instruments_performance_attribution(time_period=time_period)
        elif attribution_metric == AttributionMetric.PNL_RISK:
            data = self.get_instruments_pnl_risk_attribution(time_period=time_period)
        elif attribution_metric == AttributionMetric.INST_PNL:
            data = self.get_instruments_navs(time_period=time_period)
        else:
            raise NotImplementedError(f"{attribution_metric}")

        return data

    def get_num_investable_instruments(self, time_period: da.TimePeriod = None) -> pd.Series:
        exposures = self.weights.replace({0.0: np.nan})
        count = np.sum(np.where(np.isfinite(exposures), 1.0, 0.0), axis=1)
        num_investable_instruments = pd.Series(count, index=exposures.index, name=self.nav.name)
        if time_period is not None:
            num_investable_instruments = time_period.locate(num_investable_instruments)
        return num_investable_instruments

    def get_instruments_performance_table(self,
                                          time_period: da.TimePeriod = None,
                                          portfolio_name: str = 'Attribution'
                                          ) -> pd.DataFrame:
        """
        using avg weight
        """
        insts_returns = self.get_instruments_returns(time_period=time_period)
        insts_return = ret.to_total_returns(prices=ret.returns_to_nav(returns=insts_returns))

        weight = self.weights
        if time_period is not None:
            weight = time_period.locate(weight)
        weight = weight.mean(axis=0)
        portf_return = insts_return.multiply(weight).replace({0.0: np.nan}).dropna()
        data = pd.concat([weight.rename('Weight'),
                          insts_return.rename('Asset'),
                          portf_return.rename(portfolio_name)],
                         axis=1).dropna()
        data = data.sort_values('Weight', ascending=False)
        if self.tickers_to_names_map is not None:
            data = data.rename(index=self.tickers_to_names_map)

        return data

    def get_attribution_table_by_instrument(self,
                                            time_period: da.TimePeriod = None,
                                            freq: str = 'M',
                                            ) -> pd.DataFrame:
        """
        using avg weight
        """
        returns_f = self.get_instruments_periodic_returns(time_period=time_period, freq=freq)
        weight = self.weights.reindex(index=returns_f.index, method='ffill').shift(1)
        # first row is None
        portf_return = returns_f.multiply(weight).iloc[1:, :]
        if self.tickers_to_names_map is not None:
            portf_return = portf_return.rename(columns=self.tickers_to_names_map)
        return portf_return

    """
    plotting methods
    """
    def plot_nav(self, time_period: da.TimePeriod = None, **kwargs) -> None:
        nav = self.get_portfolio_nav(time_period=time_period)
        with sns.axes_style('darkgrid'):
            fig, axs = plt.subplots(2, 1, figsize=(16, 12), tight_layout=True)
            ppd.plot_prices_with_dd(prices=nav,
                                    axs=axs,
                                    **kwargs)

    def plot_ra_perf_table(self,
                           benchmark_price: pd.Series = None,
                           is_grouped: bool = True,
                           time_period: da.TimePeriod = None,
                           perf_params: PerfParams = None,
                           title: str = None,
                           ax: plt.Subplot = None,
                           **kwargs) -> None:
        if is_grouped:
            prices = self.get_ac_navs(time_period=time_period)
            title = title or f"RA performance table by groups: {da.get_time_period(prices).to_str()}"
        else:
            prices = self.get_portfolio_nav(time_period=time_period)
            title = title or f"RA performance table: {da.get_time_period(prices).to_str()}"
        if benchmark_price is not None:
            prices = pd.concat([prices, benchmark_price.reindex(index=prices.index, method='ffill')], axis=1)
            ppt.plot_ra_perf_table_benchmark(prices=prices,
                                             benchmark=str(benchmark_price.name),
                                             perf_params=perf_params,
                                             perf_columns=rpt.BENCHMARK_TABLE_COLUMNS,
                                             title=title,
                                             rotation_for_columns_headers=0,
                                             special_rows_colors=[(1, 'deepskyblue'), (len(prices.columns), 'lavender')],
                                             column_header='Portfolio',
                                             ax=ax,
                                             **kwargs)
        else:
            ppt.plot_ra_perf_table(prices=prices,
                                   perf_params=perf_params,
                                   perf_columns=rpt.COMPACT_TABLE_COLUMNS,
                                   title=title,
                                   rotation_for_columns_headers=0,
                                   column_header='Portfolio',
                                   ax=ax,
                                   **kwargs)

    def plot_returns_scatter(self,
                             benchmark_price: pd.Series = None,
                             is_grouped: bool = True,
                             time_period: da.TimePeriod = None,
                             title: str = None,
                             freq: str = 'Q',
                             ax: plt.Subplot = None,
                             **kwargs
                             ) -> None:
        if is_grouped:
            prices = self.get_ac_navs(time_period=time_period)
            title = title or f"Scatterplot of {freq}-returns by groups vs {str(benchmark_price.name)}"
        else:
            prices = self.get_portfolio_nav(time_period=time_period)
            title = title or f"Scatterplot of {freq}-returns vs {str(benchmark_price.name)}"
        prices = pd.concat([prices, benchmark_price.reindex(index=prices.index, method='ffill')], axis=1)
        local_kwargs = sop.update_kwargs(kwargs=kwargs,
                                         new_kwargs={'weight': 'bold',
                                                     #'alpha_an_factor': 52.0,
                                                     'x_rotation': 0,
                                                     'first_color_fixed': False,
                                                     'ci': None})
        prs.plot_returns_scatter(prices=prices,
                                 benchmark=str(benchmark_price.name),
                                 freq=freq,
                                 order=2,
                                 title=title,
                                 ax=ax,
                                 **local_kwargs)

    def plot_monthly_returns_heatmap(self, time_period: da.TimePeriod = None,
                                     heatmap_freq: str = 'A',
                                     date_format: str = '%Y',
                                     ax: plt.Subplot = None,
                                     **kwargs
                                     ) -> None:
        plot_returns_heatmap(prices=self.get_portfolio_nav(time_period=time_period),
                             heatmap_column_freq='M',
                             is_add_annual_column=True,
                             is_inverse_order=True,
                             heatmap_freq=heatmap_freq,
                             date_format=date_format,
                             ax=ax,
                             **kwargs)

    def plot_periodic_returns(self,
                              benchmark_prices: pd.DataFrame = None,
                              is_grouped: bool = True,
                              time_period: da.TimePeriod = None,
                              heatmap_freq: str = 'A',
                              date_format: str = '%Y',
                              transpose: bool = True,
                              title: str = None,
                              ax: plt.Subplot = None,
                              **kwargs
                              ) -> None:
        if is_grouped:
            prices = self.get_ac_navs(time_period=time_period)
            title = title or f"{heatmap_freq}-returns by groups"
        else:
            prices = self.get_portfolio_nav(time_period=time_period)
            title = title or f"{heatmap_freq}-returns"
        if benchmark_prices is not None:
            hline_rows = [len(prices.columns)]
            prices = pd.concat([prices, benchmark_prices.reindex(index=prices.index, method='ffill')], axis=1)
        else:
            hline_rows = None

        rhe.plot_periodic_returns_table(prices=prices,
                                                    freq=heatmap_freq,
                                                    ax=ax,
                                                    title=title,
                                                    date_format=date_format,
                                                    transpose=transpose,
                                                    hline_rows=hline_rows,
                                                    **kwargs)

    def plot_regime_data(self,
                         benchmark_price: pd.Series,
                         is_grouped: bool = True,
                         regime_data_to_plot: RegimeData = RegimeData.REGIME_SHARPE,
                         time_period: da.TimePeriod = None,
                         var_format: Optional[str] = None,
                         is_conditional_sharpe: bool = True,
                         legend_loc: Optional[str] = 'upper center',
                         title: str = None,
                         perf_params: PerfParams = None,
                         regime_params: BenchmarkReturnsQuantileRegimeSpecs = None,
                         ax: plt.Subplot = None,
                         **kwargs
                         ) -> plt.Figure:

        if is_grouped:
            prices = self.get_ac_navs(time_period=time_period)
            title = title or f"Sharpe ratio decomposition by groups to {str(benchmark_price.name)} Bear/Normal/Bull regimes"
        else:
            prices = self.get_portfolio_nav(time_period=time_period)
            title = title or f"Sharpe ratio decomposition to {str(benchmark_price.name)} Bear/Normal/Bull regimes"
        prices = pd.concat([benchmark_price.reindex(index=prices.index, method='ffill'), prices], axis=1)

        regime_classifier = rcl.BenchmarkReturnsQuantilesRegime(regime_params=regime_params)
        fig = qis.plots.derived.regime_data.plot_regime_data(regime_classifier=regime_classifier,
                                                             prices=prices,
                                                             benchmark=str(benchmark_price.name),
                                                             is_conditional_sharpe=is_conditional_sharpe,
                                                             regime_data_to_plot=regime_data_to_plot,
                                                             var_format=var_format or '{:.2f}',
                                                             legend_loc=legend_loc,
                                                             perf_params=perf_params,
                                                             title=title,
                                                             ax=ax,
                                                             **kwargs)
        return fig

    def plot_vol_regimes(self,
                         benchmark_price: pd.Series,
                         is_grouped: bool = True,
                         time_period: da.TimePeriod = None,
                         title: str = None,
                         regime_params: BenchmarkReturnsQuantileRegimeSpecs = None,
                         ax: plt.Subplot = None,
                         **kwargs
                         ) -> plt.Figure:

        if is_grouped:
            prices = self.get_ac_navs(time_period=time_period)
            title = title or f"{regime_params.freq}-returns by groups conditional on vols {str(benchmark_price.name)}"
        else:
            prices = self.get_portfolio_nav(time_period=time_period)
            title = title or f"{regime_params.freq}-returns conditional on vols {str(benchmark_price.name)}"
        prices = pd.concat([benchmark_price.reindex(index=prices.index, method='ffill'), prices], axis=1)

        regime_classifier = rcl.BenchmarkVolsQuantilesRegime(regime_params=rcl.VolQuantileRegimeSpecs(freq=regime_params.freq))
        fig = qis.plots.derived.regime_data.plot_regime_boxplot(regime_classifier=regime_classifier,
                                                                prices=prices,
                                                                benchmark=str(benchmark_price.name),
                                                                title=title,
                                                                ax=ax,
                                                                **kwargs)
        return fig

    def plot_contributors(self,
                          time_period: da.TimePeriod = None,
                          num_assets: int = 10,
                          ax: plt.Subplot = None,
                          **kwargs
                          ) -> None:
        prices = self.get_instruments_navs(time_period=time_period)
        ppt.plot_top_bottom_performers(prices=prices, num_assets=num_assets, ax=ax, **kwargs)

    def plot_pnl(self, time_period: da.TimePeriod = None) -> None:
        avg_costs, realized_pnl, mtm_pnl, total_pnl, trades = self.compute_realized_pnl(time_period=time_period)
        prices = self.prices
        if time_period is not None:
            prices = time_period.locate(prices)
        with sns.axes_style('darkgrid'):
            fig, axs = plt.subplots(5, 1, figsize=(10, 16), tight_layout=True)
            pts.plot_time_series(df=prices, legend_stats=pts.LegendStats.FIRST_AVG_LAST, title='prices', ax=axs[0])
            pts.plot_time_series(df=avg_costs, legend_stats=pts.LegendStats.FIRST_AVG_LAST, title='avg_costs', ax=axs[1])
            pts.plot_time_series(df=realized_pnl, legend_stats=pts.LegendStats.FIRST_AVG_LAST, title='realized_pnl', ax=axs[2])
            pts.plot_time_series(df=mtm_pnl, legend_stats=pts.LegendStats.FIRST_AVG_LAST, title='mtm_pnl', ax=axs[3])
            pts.plot_time_series(df=total_pnl, legend_stats=pts.LegendStats.FIRST_AVG_LAST, title='total_pnl', ax=axs[4])

    def plot_weights(self,
                     is_input_weights: bool = True,
                     columns: List[str] = None,
                     freq: str = 'W-WED',
                     bbox_to_anchor: Tuple[float, float] = (0.4, 1.14),
                     ax: plt.Subplot = None,
                     title: str = '',
                     **kwargs
                     ) -> None:
        if is_input_weights:
            weights = self.input_weights.copy()
        else:
            weights = self.weights.copy()
        if columns is not None:
            weights = weights[columns]
        weights = weights.resample(freq).last().fillna(method='ffill')
        pst.plot_stack(df=weights,
                       add_mean_levels=False,
                       is_use_bar_plot=True,
                       # is_yaxis_limit_01=True,
                       baseline='zero',
                       bbox_to_anchor=bbox_to_anchor,
                       title=title,
                       legend_stats=pst.LegendStats.FIRST_AVG_LAST,
                       var_format='{:.1%}',
                       ax=ax,
                       **kwargs)


@njit
def compute_realized_pnl(prices: np.ndarray,
                         units: np.ndarray
                         ) -> Tuple[np.ndarray, ...]:
    """
    pnl for long only positions, computes avg entry price and pnl by instrument
    """
    avg_costs = np.zeros_like(prices)
    realized_pnl = np.zeros_like(prices)
    mtm_pnl = np.zeros_like(prices)
    trades = np.zeros_like(prices)
    for idx, (price1, unit1) in enumerate(zip(prices, units)):
        if idx == 0:
            avg_costs[idx] = np.where(np.greater(unit1, 1e-16), price1, 0.0)
        else:
            unit0 = units[idx-1]
            avg_costs0 = avg_costs[idx-1]
            delta = unit1 - unit0
            is_purchase = np.greater(delta, 1e-16)
            is_sell = np.less(delta, -1e-16)
            realized_pnl[idx] = np.where(is_sell, -delta*(price1-avg_costs0), 0.0)
            avg_costs[idx] = np.where(is_purchase, np.true_divide(delta*price1+unit0*avg_costs0, unit1), avg_costs0)
            mtm_pnl[idx] = unit0*(price1 - avg_costs0) - realized_pnl[idx]
            trades[idx] = delta
    return avg_costs, realized_pnl, mtm_pnl, trades


class AllocationType(EnumMap):
    EW = 1
    FIXED_WEIGHTS = 2
    ERC = 3
    ERC_ALT = 4


@dataclass
class PortfolioInput:
    """
    define data inputs for portfolio construction
    """
    name: str
    weights: Union[np.ndarray, pd.DataFrame, Dict[str, float]]
    prices: pd.DataFrame = None  # mandatory but we set none for enumarators
    allocation_type: AllocationType = AllocationType.FIXED_WEIGHTS
    time_period: da.TimePeriod = None
    rebalance_freq: str = 'Q'
    regime_freq: str = 'M'
    returns_freq: str = 'M'
    ewm_lambda: float = 0.92
    target_vol: float = None

    def update(self, new: Dict[Any, Any]):
        for key, value in new.items():
            if hasattr(self, key):
                setattr(self, key, value)



"""
path management for saving and reading from local drive
use settings.yaml to set local directories

Content of settings.yaml:
RESOURCE_PATH:
  'C:\\your_folder\\'
UNIVERSE_PATH:
  'C:\\your_folder\\'
OUTPUT_PATH:
  'C:\\your_folder\\'
# optional
POSTGRES:
  "postgresql://user:password@database:port"
"""
import datetime
import functools
import platform
import time
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from os import listdir
from os.path import isfile, join
from typing import Dict, List, NamedTuple, Optional, Union
from matplotlib.backends.backend_pdf import PdfPages
from sqlalchemy.engine.base import Engine
from enum import Enum


""""
Path specifications
"""
from qis.local_path import get_paths
LOCAL_PATHS = get_paths()
RESOURCE_PATH = LOCAL_PATHS['RESOURCE_PATH']
UNIVERSE_PATH = LOCAL_PATHS['UNIVERSE_PATH']
OUTPUT_PATH = LOCAL_PATHS['OUTPUT_PATH']


DATE_FORMAT = '%Y%m%d_%H%M'
INDEX_COLUMN = 'timestamp'  # for Postresql


class FileData(NamedTuple):
    extension: str
    folder: Union[str, None]


class FileTypes(FileData, Enum):
    FIGURE = FileData(extension='.png', folder='figures')
    PNG = FileData(extension='.PNG', folder='figures')
    EPS = FileData(extension='.eps', folder='figures')
    SVG = FileData(extension='.svg', folder='figures')
    CSV = FileData(extension='.csv', folder='csv')
    FEATHER = FileData(extension='.feather', folder='feather')
    EXCEL = FileData(extension='.xlsx', folder='excel')
    PPTX = FileData(extension='.pptx', folder='pptx')
    WORD = FileData(extension='.docx', folder=None)
    PDF = FileData(extension='.pdf', folder=None)
    TXT = FileData(extension='.txt', folder='txt')
    PARQUET = FileData(extension='.parquet', folder='parquet')
    PARAMS = FileData(extension='.xlsx', folder='data\\params')
    ZIP = FileData(extension='.zip', folder=None)


class PathData(NamedTuple):
    platform: str
    path: str


class ResourcePath(PathData, Enum):
    """
    specify window and lynox local paths to read resourse files
    """
    WINDOWS = PathData(platform='Windows', path=RESOURCE_PATH)
    LINUX = PathData(platform='Linux', path=RESOURCE_PATH)


class OutputPath(PathData, Enum):
    """
    specify window and lynox local paths for outputs
    """
    WINDOWS = PathData(platform='Windows', path=OUTPUT_PATH)
    LINUX = PathData(platform='Linux', path=OUTPUT_PATH)


def get_output_path() -> str:
    """
    platform dependent output path
    """
    run_platform = platform.system()
    path_data = next((item for item in OutputPath if item.value.platform == run_platform), None)
    if path_data is None:
        raise TypeError(f"unknown platform {run_platform}")
    return path_data.path


def get_resource_path() -> str:
    """
    platform dependent resourse path
    """
    run_platform = platform.system()
    path_data = next((item for item in ResourcePath if item.value.platform == run_platform), None)
    if path_data is None:
        raise TypeError(f"unknown platform {run_platform}")
    return path_data.path


def join_file_name_parts(parts: List[str]) -> str:
    """
    set standard for joining parts of file name
    """
    return '_'.join(parts)


def get_local_file_path(file_name: Optional[str],
                        file_type: Optional[FileTypes] = FileTypes.CSV,
                        local_path: Optional[str] = None,
                        subfolder_name: str = None,
                        subsubfolder_name: str = None,
                        key: str = None,
                        is_output_file: bool = False
                        ) -> str:
    """
    file data management is organised as:
    RESOURCE_PATH/subfolder_name/subsubfolder_name/file_name+key.file_type.value
    default value without optional arguments will be:
    RESOURCE_PATH/file_name.file_type.value
    the default path is resource_file_path/
    """
    if local_path is None:
        if file_name is None and key is None and subfolder_name is None and subsubfolder_name is None:
            raise ValueError(f"file_name or key or subfolder_name must be given")

        if is_output_file:
            local_path = OUTPUT_PATH
        else:
            local_path = RESOURCE_PATH

        if subfolder_name is not None:
            if subsubfolder_name is not None:
                local_path = join(local_path, subfolder_name, subsubfolder_name)
            else:
                local_path = join(local_path, subfolder_name)

    if key is not None:
        if file_name is None:
            file_name = key
        else:
            file_name = join_file_name_parts([file_name, key])
    else:
        if file_name is None:  # will return directory path
            file_name = ''

    if file_name is not None:
        local_path = join(local_path, file_name)

    if file_type is not None:  # by convention no file folder
        file_path = local_path + file_type.extension
    else:
        file_path = local_path

    return file_path


def get_param_file_path(file_name: Optional[str] = None,
                        file_type: FileTypes = FileTypes.PARAMS
                        ) -> str:
    path_data = RESOURCE_PATH
    if file_name is not None:
        file_path = join(path_data, file_type.folder, file_name + file_type.extension)
    else:
        file_path = join(path_data, file_type.folder)
    return file_path


def timer(func):
    """
    Print the runtime of the decorated function
    """
    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start_time = time.perf_counter()    # 1
        value = func(*args, **kwargs)
        end_time = time.perf_counter()      # 2
        run_time = end_time - start_time    # 3
        if run_time < 60.0:
            print(f"Finished {func.__name__!r} in {run_time:.4f} secs")
        else:
            minuts = np.floor(run_time/60.0)
            secs = run_time - 60.0*minuts
            print(f"Finished {func.__name__!r} in {minuts:.0f}m {secs:.0f}secs")
        return value
    return wrapper_timer


"""
Pandas to/from Excel core
"""


def delocalize_df(data: pd.DataFrame, delocalize: bool) -> pd.DataFrame:
    if delocalize and isinstance(data.index, pd.DatetimeIndex):
        data.index = data.index.tz_localize(None)
    return data


def save_df_to_excel(data: Union[pd.DataFrame, List[pd.DataFrame], Dict[str, pd.DataFrame]],
                     file_name: str,
                     local_path: Optional[str] = None,
                     subfolder_name: str = None,
                     subsubfolder_name: str = None,
                     key: str = None,
                     add_current_date: bool = True,
                     sheet_names: List[str] = None,
                     transpose: bool = False,
                     is_output_file: bool = False
                     ) -> str:
    """
    pandas or list of pandas to one excel
    """
    if is_output_file and add_current_date:
        file_name = f"{file_name}_{datetime.datetime.now().strftime(DATE_FORMAT)}"

    file_path = get_local_file_path(file_name=file_name,
                                    file_type=FileTypes.EXCEL,
                                    local_path=local_path,
                                    subfolder_name=subfolder_name,
                                    subsubfolder_name=subsubfolder_name,
                                    key=key,
                                    is_output_file=is_output_file)

    excel_writer = pd.ExcelWriter(file_path)
    if isinstance(data, list):  # publish with sheet names
        if sheet_names is None:
            sheet_names = [f"Sheet {n+1}" for n, _ in enumerate(data)]
        for df, name in zip(data, sheet_names):
            df = delocalize_df(df, delocalize=True)
            if transpose:
                df = df.T
            df.to_excel(excel_writer=excel_writer, sheet_name=name)
    elif isinstance(data, dict):  # publish with sheet names
        for key, df in data.items():
            df = delocalize_df(df, delocalize=True)
            if transpose:
                df = df.T
            df.to_excel(excel_writer=excel_writer, sheet_name=key)
    else:
        if transpose:
            data = data.T
        data = delocalize_df(data, delocalize=True)
        data.to_excel(excel_writer=excel_writer)

    excel_writer.close()

    return file_path


def load_df_from_excel(file_name: str,
                       sheet_name: str = 'Sheet1',
                       local_path: Optional[str] = None,
                       subfolder_name: str = None,
                       subsubfolder_name: str = None,
                       key: str = None,
                       is_index: bool = True,
                       delocalize: bool = False  # excel data may have local time which are unwanted
                       ) -> pd.DataFrame:
    """
    one file, one sheet to pandas
    """
    file_path = get_local_file_path(file_name=file_name,
                                    file_type=FileTypes.EXCEL,
                                    local_path=local_path,
                                    subfolder_name=subfolder_name,
                                    subsubfolder_name=subsubfolder_name,
                                    key=key)
    try:
        excel_reader = pd.ExcelFile(file_path, engine='openpyxl')
    except FileNotFoundError:
        raise TypeError(f"file data {file_path} nor found")

    index_col = 0 if is_index else None
    df = excel_reader.parse(sheet_name=sheet_name, index_col=index_col)

    if is_index:
        df = delocalize_df(df, delocalize=delocalize)

    return df


def save_df_dict_to_excel(datasets: Dict[Union[str, Enum, NamedTuple], pd.DataFrame],
                          file_name: str,
                          local_path: Optional[str] = None,
                          subfolder_name: str = None,
                          subsubfolder_name: str = None,
                          key: str = None,
                          add_current_date: bool = False,
                          is_output_file: bool = False,
                          delocalize: bool = False
                          ) -> str:
    """
    dictionary of pandas to same Excel
    """
    if is_output_file and add_current_date:
        file_name = f"{file_name}_{datetime.datetime.now().strftime(DATE_FORMAT)}"

    file_path = get_local_file_path(file_name=file_name,
                                    file_type=FileTypes.EXCEL,
                                    local_path=local_path,
                                    subfolder_name=subfolder_name,
                                    subsubfolder_name=subsubfolder_name,
                                    key=key,
                                    is_output_file=is_output_file)

    excel_writer = pd.ExcelWriter(file_path)
    for key, data in datasets.items():
        data = delocalize_df(data, delocalize=delocalize)
        data.to_excel(excel_writer=excel_writer, sheet_name=key)
    excel_writer.close()
    return file_path


def load_df_dict_from_excel(file_name: str,
                            dataset_keys: Optional[List[Union[str, Enum, NamedTuple]]],
                            local_path: Optional[str] = None,
                            subfolder_name: str = None,
                            subsubfolder_name: str = None,
                            key: str = None,
                            is_index: bool = True,
                            delocalize: bool = False,
                            tz: str = None
                            ) -> Dict[str, pd.DataFrame]:
    """
    loag Excel sheets to pandas dict
    dataset_keys = None, it will read all sheets
    """
    file_path = get_local_file_path(file_name=file_name,
                                    file_type=FileTypes.EXCEL,
                                    local_path=local_path,
                                    subfolder_name=subfolder_name,
                                    subsubfolder_name=subsubfolder_name,
                                    key=key)
    try:
        excel_reader = pd.ExcelFile(file_path, engine='openpyxl')
    except FileNotFoundError:
        raise TypeError(f"file data {file_path} not found")
    if dataset_keys is None:
        dataset_keys = excel_reader.sheet_names
    index_col = 0 if is_index else None
    pandas_dict = {}
    for key in dataset_keys:
        try:
            df = excel_reader.parse(sheet_name=key, index_col=index_col)
        except:
            raise TypeError(f"sheet_name data {key} nor found")
        df = delocalize_df(df, delocalize=delocalize)
        if tz is not None:
            df.index = df.index.tz_localize(tz)
        pandas_dict[key] = df

    return pandas_dict


"""
Pandas to/from CSV core
"""


def save_df_to_csv(df: pd.DataFrame,
                   file_name: str = None,
                   subfolder_name: str = None,
                   subsubfolder_name: str = None,
                   key: str = None,
                   local_path: Optional[str] = None,
                   delocalize: bool = False,
                   is_output_file: bool = False
                   ) -> None:
    """
    pandas to csv
    """
    file_path = get_local_file_path(file_name=file_name,
                                    file_type=FileTypes.CSV,
                                    local_path=local_path,
                                    subfolder_name=subfolder_name,
                                    subsubfolder_name=subsubfolder_name,
                                    key=key,
                                    is_output_file=is_output_file)
    df = delocalize_df(df, delocalize=delocalize)
    df.to_csv(path_or_buf=file_path)


def load_df_from_csv(file_name: Optional[str] = None,
                     local_path: Optional[str] = None,
                     subfolder_name: str = None,
                     subsubfolder_name: str = None,
                     key: str = None,
                     is_index: bool = True,
                     parse_dates: bool = True,
                     delocalize: bool = False,  # excel data may have local time which are unwanted
                     dayfirst: Optional[bool] = None,
                     tz: str = None,
                     drop_duplicated: bool = False
                     ) -> pd.DataFrame:
    """
    pandas from csv
    """
    file_path = get_local_file_path(file_name=file_name,
                                    file_type=FileTypes.CSV,
                                    local_path=local_path,
                                    subfolder_name=subfolder_name,
                                    subsubfolder_name=subsubfolder_name,
                                    key=key)

    if is_index:
        index_col = 0
        parse_dates = parse_dates
    else:
        index_col = None
        parse_dates = None

    try:
        df = pd.read_csv(filepath_or_buffer=file_path,
                         index_col=index_col,
                         parse_dates=parse_dates,
                         dayfirst=dayfirst)

    except UnicodeDecodeError:  # try without index
        df = pd.read_csv(filepath_or_buffer=file_path)

    except FileNotFoundError:
        raise FileNotFoundError(f"not found {file_name} with file_path={file_path}")

    if drop_duplicated:
        df = df.loc[~df.index.duplicated(keep='first')]

    if not df.empty:
        if tz is not None:
            if df.index.tzinfo is not None:
                df.index = df.index.tz_convert(tz=tz)
            else:
                df.index = df.index.tz_localize(tz=tz)
        elif delocalize:
            df = delocalize_df(df, delocalize=delocalize)

    return df


def save_df_dict_to_csv(datasets: Dict[Union[str, Enum, NamedTuple], pd.DataFrame],
                        file_name: Optional[str] = None,
                        local_path: Optional[str] = None,
                        subfolder_name: str = None,
                        subsubfolder_name: str = None,
                        delocalize: bool = False,
                        is_output_file: bool = False
                        ) -> None:
    """
    pandas dict to csv files
    """
    for key, data in datasets.items():
        if data is not None:
            file_path = get_local_file_path(file_name=file_name,
                                            file_type=FileTypes.CSV,
                                            local_path=local_path,
                                            subfolder_name=subfolder_name,
                                            subsubfolder_name=subsubfolder_name,
                                            key=key,
                                            is_output_file=is_output_file)
            data = delocalize_df(data, delocalize=delocalize)
            data.to_csv(path_or_buf=file_path)


def load_df_dict_from_csv(dataset_keys: List[Union[str, Enum, NamedTuple]],
                          file_name: Optional[str],
                          local_path: Optional[str] = None,
                          subfolder_name: str = None,
                          subsubfolder_name: str = None,
                          is_index: bool = True,
                          force_not_found_error: bool = False,
                          delocalize: bool = False
                          ) -> Dict[str, pd.DataFrame]:
    """
    pandas dict from csv files
    """
    index_col = 0 if is_index else None
    pandas_dict = {}
    for key in dataset_keys:
        file_path = get_local_file_path(file_name=file_name,
                                        file_type=FileTypes.CSV,
                                        local_path=local_path,
                                        subfolder_name=subfolder_name,
                                        subsubfolder_name=subsubfolder_name,
                                        key=key)
        try:
            data = pd.read_csv(filepath_or_buffer=file_path,
                               index_col=index_col,
                               parse_dates=True)
            data = delocalize_df(data, delocalize=delocalize)
            pandas_dict[key] = data
        except:
            message = f"file data {file_path}, {key} not found"
            if force_not_found_error:
                raise FileNotFoundError(message)
            else:
                print(message)

    return pandas_dict


#############################################################
###  Pandas to/from feather
#############################################################

@timer
def save_df_dict_to_sql(engine: Engine,
                        table_name: str,
                        dfs: Dict[Union[str, Enum, NamedTuple], pd.DataFrame],
                        schema: Optional[str] = None,
                        index_col: Optional[str] = INDEX_COLUMN
                        ) -> None:
    """
    save pandas dict to sql engine
    """
    for key, df in dfs.items():
        if df is not None:
            if index_col is not None:
                df = df.reset_index(names=index_col)
            df.to_sql(f"{table_name}_{key}", engine, schema=schema, if_exists='replace')


@timer
def load_df_dict_from_sql(engine: Engine,
                          table_name: str,
                          dataset_keys: List[Union[str, Enum, NamedTuple]],
                          schema: Optional[str] = None,
                          index_col: Optional[str] = INDEX_COLUMN,
                          columns: Optional[List[str]] = None
                          ) -> Dict[str, pd.DataFrame]:
    """
    pandas dict from csv files
    """
    pandas_dict = {}
    for key in dataset_keys:
        df = pd.read_sql_table(table_name=f"{table_name}_{key}", con=engine, schema=schema,
                               index_col=index_col,
                               columns=columns)
        if index_col is not None and index_col in df.columns:
            df[index_col] = pd.to_datetime(df[index_col])
            df = df.set_index(index_col)
        pandas_dict[key] = df
    return pandas_dict


#############################################################
###  Pandas to/from feather
#############################################################

def save_df_to_feather(df: pd.DataFrame,
                       file_name: Optional[str] = None,
                       local_path: Optional[str] = None,
                       subfolder_name: str = None,
                       subsubfolder_name: str = None,
                       is_output_file: bool = False,
                       index_col: Optional[str] = INDEX_COLUMN
                       ) -> None:
    """
    pandas dict to feather files
    """
    file_path = get_local_file_path(file_name=file_name,
                                    file_type=FileTypes.FEATHER,
                                    local_path=local_path,
                                    subfolder_name=subfolder_name,
                                    subsubfolder_name=subsubfolder_name,
                                    is_output_file=is_output_file)
    if index_col not in df.columns:
        df = df.reset_index(names=index_col)
    df.to_feather(path=file_path)


def load_df_from_feather(file_name: Optional[str] = None,
                         local_path: Optional[str] = None,
                         subfolder_name: str = None,
                         subsubfolder_name: str = None,
                         key: str = None,
                         index_col: Optional[str] = INDEX_COLUMN
                         ) -> pd.DataFrame:
    """
    load dfs from feather files
    """
    file_path = get_local_file_path(file_name=file_name,
                                    file_type=FileTypes.FEATHER,
                                    local_path=local_path,
                                    subfolder_name=subfolder_name,
                                    subsubfolder_name=subsubfolder_name,
                                    key=key)
    try:
        df = pd.read_feather(file_path)
    except:
        raise FileNotFoundError(f"not found {file_name} with file_path={file_path}")

    if index_col is not None and index_col in df.columns:
        df[index_col] = pd.to_datetime(df[index_col])
        df = df.set_index(index_col)
    return df


@timer
def save_df_dict_to_feather(dfs: Dict[Union[str, Enum, NamedTuple], pd.DataFrame],
                            file_name: Optional[str] = None,
                            local_path: Optional[str] = None,
                            subfolder_name: str = None,
                            subsubfolder_name: str = None,
                            is_output_file: bool = False
                            ) -> None:
    """
    pandas dict to csv files
    """
    for key, df in dfs.items():
        if df is not None:
            file_path = get_local_file_path(file_name=file_name,
                                            file_type=FileTypes.FEATHER,
                                            local_path=local_path,
                                            subfolder_name=subfolder_name,
                                            subsubfolder_name=subsubfolder_name,
                                            key=key,
                                            is_output_file=is_output_file)
            df = df.reset_index(names='index')
            df.to_feather(path=file_path)


def load_df_dict_from_feather(dataset_keys: List[Union[str, Enum, NamedTuple]],
                              file_name: Optional[str],
                              local_path: Optional[str] = None,
                              subfolder_name: str = None,
                              subsubfolder_name: str = None,
                              index_col: Optional[str] = INDEX_COLUMN,
                              force_not_found_error: bool = False,
                              ) -> Dict[str, pd.DataFrame]:
    """
    pandas dict from csv files
    """
    pandas_dict = {}
    for key in dataset_keys:
        file_path = get_local_file_path(file_name=file_name,
                                        file_type=FileTypes.FEATHER,
                                        local_path=local_path,
                                        subfolder_name=subfolder_name,
                                        subsubfolder_name=subsubfolder_name,
                                        key=key)
        try:
            df = pd.read_feather(file_path)
            if index_col is not None and index_col in df.columns:
                df.index = pd.to_datetime(df[index_col])
                df = df.set_index(index_col)
            pandas_dict[key] = df
        except:
            message = f"file data {file_path}, {key} not found"
            if force_not_found_error:
                raise FileNotFoundError(message)
            else:
                print(message)

    return pandas_dict


"""
Pandas to/from parquet core
"""


def save_df_to_parquet(df: pd.DataFrame,
                       file_name: str,
                       subfolder_name: Optional[str] = None,
                       subsubfolder_name: str = None,
                       key: Optional[str] = None,
                       local_path: Optional[str] = None,
                       is_output_file: bool = False,
                       delocalize: bool = False
                       ) -> None:
    """
    pandas to parquet
    """
    file_path = get_local_file_path(file_name=file_name,
                                    file_type=FileTypes.PARQUET,
                                    local_path=local_path,
                                    subfolder_name=subfolder_name,
                                    subsubfolder_name=subsubfolder_name,
                                    key=key,
                                    is_output_file=is_output_file)
    df = delocalize_df(df, delocalize=delocalize)
    df.to_parquet(path=file_path)


def load_df_from_parquet(file_name: Optional[str],
                         subfolder_name: str = None,
                         subsubfolder_name: str = None,
                         key: Optional[str] = None,
                         local_path: Optional[str] = None,
                         delocalize: bool = False
                         ) -> pd.DataFrame:
    """
    pandas from parquet
    """
    file_path = get_local_file_path(file_name=file_name,
                                    file_type=FileTypes.PARQUET,
                                    local_path=local_path,
                                    subfolder_name=subfolder_name,
                                    subsubfolder_name=subsubfolder_name,
                                    key=key)
    try:
        df = pd.read_parquet(path=file_path)
    except:
        raise FileNotFoundError(f"not found {file_name} with path={file_path}")

    df = delocalize_df(df, delocalize=delocalize)

    return df


def save_df_dict_to_parquet(datasets: Dict[Union[str, Enum, NamedTuple], pd.DataFrame],
                            file_name: Optional[str] = None,
                            subfolder_name: str = None,
                            subsubfolder_name: str = None,
                            local_path: Optional[str] = None,
                            delocalize: bool = False,
                            is_output_file: bool = False
                            ) -> None:
    """
    pandas dict to parquet files
    """
    for key, data in datasets.items():
        if data is not None:
            file_path = get_local_file_path(file_name=file_name,
                                            file_type=FileTypes.PARQUET,
                                            local_path=local_path,
                                            subfolder_name=subfolder_name,
                                            subsubfolder_name=subsubfolder_name,
                                            key=key,
                                            is_output_file=is_output_file)
            data = delocalize_df(data, delocalize=delocalize)
            data.to_parquet(path=file_path)


def load_df_dict_from_parquet(dataset_keys: List[Union[str, Enum, NamedTuple]],
                              file_name: Optional[str],
                              subfolder_name: str = None,
                              subsubfolder_name: str = None,
                              local_path: Optional[str] = None,
                              force_not_found_error: bool = False
                              ) -> Dict[str, pd.DataFrame]:
    """
    pandas dict from parquet files
    """
    pandas_dict = {}
    for key in dataset_keys:
        file_path = get_local_file_path(file_name=file_name,
                                        file_type=FileTypes.PARQUET,
                                        local_path=local_path,
                                        subfolder_name=subfolder_name,
                                        subsubfolder_name=subsubfolder_name,
                                        key=key)
        try:
            pandas_dict[key] = pd.read_parquet(path=file_path)

        except:
            message = f"file data {file_name}, {key} not found"
            if force_not_found_error:
                raise FileNotFoundError(message)
            else:
                print(message)

    return pandas_dict


"""
For pdfs
"""


def get_pdf_path(file_name: str,
                 local_path: Union[None, str] = None,
                 add_current_date: bool = True
                 ) -> str:

    if add_current_date:
        file_name = join_file_name_parts([file_name, datetime.datetime.now().strftime(DATE_FORMAT)])

    file_path = get_local_file_path(file_name=file_name, file_type=FileTypes.PDF, local_path=local_path, is_output_file=True)

    return file_path


def get_all_folder_files(folder_path: str):
    files = [f for f in listdir(folder_path) if isfile(join(folder_path, f))]
    print(files)



"""
Figure files
"""


def save_fig(fig: plt.Figure,
             file_name: str,
             local_path: Optional[str] = None,
             dpi: int = 300,
             file_type=FileTypes.PNG,
             add_current_date: bool = False,
             **kwargs
             ) -> str:
    """
    save matplotlib figure
    """
    if add_current_date:
        file_name = join_file_name_parts([file_name, datetime.datetime.now().strftime(DATE_FORMAT)])
    file_path = get_local_file_path(file_name=file_name,
                                    file_type=file_type,
                                    local_path=local_path,
                                    is_output_file=True)
    if file_type == FileTypes.PNG:
        fig.savefig(file_path, dpi=dpi)
    elif file_type == FileTypes.EPS:
        fig.savefig(file_path, dpi=dpi, format='eps')
    elif file_type == FileTypes.SVG:
        fig.savefig(file_path, dpi=dpi, format='svg')
    elif file_type == FileTypes.PDF:
        fig.savefig(file_path, dpi=dpi, format='pdf')
    else:
        raise NotImplementedError(f"{file_type}")
    return file_path


def save_figs(figs: Dict[str, plt.Figure],
              local_path: Optional[str] = None,
              dpi: int = 300,
              file_type=FileTypes.PNG,
              add_current_date: bool = False,
              **kwargs
              ) -> None:
    """
    save matplotlib figures dict
    """
    for key, fig in figs.items():
        file_path = save_fig(fig=fig,
                             file_name=key,
                             local_path=local_path,
                             dpi=dpi,
                             file_type=file_type,
                             add_current_date=add_current_date,
                             **kwargs)
        print(file_path)


def save_figs_to_pdf(figs: Union[List[plt.Figure], Dict[str, plt.Figure]],
                     file_name: str,
                     orientation: str = 'portrait',
                     local_path: Optional[str] = None,
                     add_current_date: bool = True
                     ) -> str:
    """
    create PDF of list of plf figures
    """
    if add_current_date:
        file_name = join_file_name_parts([file_name, datetime.datetime.now().strftime(DATE_FORMAT)])

    file_path = get_local_file_path(file_name=file_name, file_type=FileTypes.PDF, is_output_file=True, local_path=local_path)

    with PdfPages(file_path) as pdf:
        if isinstance(figs, Dict):
            for _, fig in figs.items():
                if fig is not None:
                    pdf.savefig(fig, orientation=orientation)
        else:
            for fig in figs:
                if fig is not None:
                    pdf.savefig(fig, orientation=orientation)

    print(f"created PDF doc: {file_path}")
    return file_path


class UnitTests(Enum):
    LOCAL_PATHS = 1
    FOLDER_FILES = 2
    NAMES = 3
    DATA_FILE = 4
    UNIVERSE = 5


def run_unit_test(unit_test: UnitTests):

    if unit_test == UnitTests.LOCAL_PATHS:
        print(get_paths())
        print(platform.system())
        print(OUTPUT_PATH)

    elif unit_test == UnitTests.FOLDER_FILES:
        get_all_folder_files(folder_path="C://")

    elif unit_test == UnitTests.NAMES:
        file_name = join_file_name_parts(['head', 'tails'])
        print(file_name)

    elif unit_test == UnitTests.DATA_FILE:
        file_path = get_local_file_path(file_name='ETH', file_type=FileTypes.CSV)
        print(file_path)

    elif unit_test == UnitTests.UNIVERSE:
        file_path = get_local_file_path(file_name='ETH',
                                        subfolder_name='bitmex',
                                        key='1d',
                                        file_type=FileTypes.CSV)
        print(file_path)


if __name__ == '__main__':

    unit_test = UnitTests.UNIVERSE

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)



# packages
import numpy as np
import pandas as pd
import matplotlib as mpl
import matplotlib.pyplot as plt
import seaborn as sns
from typing import List, Optional, Tuple

# qis
import qis.plots.utils as put
from qis.plots.utils import LegendStats


def plot_stack(df: pd.DataFrame,
               is_use_bar_plot: bool = False,
               is_yaxis_limit_01: bool = False,
               add_mean_levels: bool = False,
               add_cum_levels: bool = False,
               add_total_line: bool = False,
               colors: List[str] = None,
               step: Optional[str] = None,  # 'mid
               title: Optional[str] = None,
               baseline: str = 'zero',  # "zero", "sym", "wiggle", "weighted_wiggle"
               ncol: int = 1,
               legend_loc: Optional[str] = 'upper center',
               legend_labels: Optional[List[str]] = None,
               legend_stats: LegendStats = LegendStats.NONE,
               var_format: str = '{:.0%}',
               fontsize: int = 10,
               linewidth: float = 1.5,
               x_date_freq: str = 'A',
               x_rotation: int = 90,
               reversed: bool = False,
               date_format: str = '%b-%y',
               bbox_to_anchor: Optional[Tuple[float, float]] = None,
               xlabel: str = None,
               ylabel: str = None,
               ax: plt.Subplot = None,
               **kwargs
               ) -> plt.Figure:

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None
    re_indexed_data, datalables = put.map_dates_index_to_str(data=df,
                                                             x_date_freq=x_date_freq,
                                                             date_format=date_format)

    if colors is None:
        colors = put.get_n_colors(n=len(re_indexed_data.columns))

    if is_use_bar_plot:  # plot bar apperas to look better for unconstraint plots
        re_indexed_data.plot.bar(stacked=True,
                                 color=colors, width=1.0, alpha=1.0, edgecolor='none',
                                 linewidth=0, ax=ax)
    else:
        ax.stackplot(re_indexed_data.index, re_indexed_data.T,
                     labels=re_indexed_data.columns, step=step, colors=colors, baseline=baseline)

    # set x axes to nearest years
    ax.set_xlim(re_indexed_data.index[0], re_indexed_data.index[-1])

    if add_total_line:  # add total as line
        totals = re_indexed_data.sum(1)
        sns.lineplot(x=re_indexed_data.index, y=totals, marker='None', color='black', ax=ax)
        # legend_labels.append('Total')
        colors.append('black')

    # change axes labels, positions of each tick, relative to the indices of the x-values
    current_ticks = ax.get_xticks()
    ax.set_xticks(np.linspace(current_ticks[0], current_ticks[-1], len(datalables)))
    ax.set_xticklabels(datalables, rotation=90, fontsize=fontsize)

    if is_yaxis_limit_01:
        ax.set_ylim(0, 1)

    if add_mean_levels or add_cum_levels:
        xmin, xmax = ax.get_xlim()
        ymin, ymax = ax.get_ylim()
        cum_mean = 0.0
        cum_mean0 = 0.0
        handles, labels = ax.get_legend_handles_labels()

        for (idx, column), handle, label in zip(enumerate(re_indexed_data.columns), handles, labels):
            mean = np.mean(re_indexed_data[column].values)
            cum_mean = cum_mean + mean

            if add_mean_levels:
                vlabel = var_format.format(mean)
            else:
                vlabel = var_format.format(cum_mean)

            if add_mean_levels and is_yaxis_limit_01 is False:  # show absolute effect
                y_loc = mean
            else: # show cumulative effect
                if column == re_indexed_data.columns[-1]:  # make it vidsiblae
                    y_loc = 1.0 - cum_mean0 if cum_mean0 < 0.5 else cum_mean0 + 0.5 * mean
                else:
                    y_loc = cum_mean
            cum_mean0 = cum_mean

            ax.axhline(y_loc, color='black', linestyle='--', linewidth=linewidth)

            color = mpl.colors.to_rgb(handle.get_facecolors()[0])
            ax.annotate(text=f"{label}={vlabel}", xy=(xmax, y_loc), fontsize=fontsize, weight='normal', color=color)

        y_annotation = 'Avg' if add_mean_levels else 'Total'
        ax.annotate(y_annotation, xy=(xmax, ymax), xytext=(1, 2), fontsize=fontsize, weight='normal',
                    textcoords='offset points', ha='left', va='bottom')

    put.set_ax_ticks_format(ax=ax, fontsize=fontsize, xvar_format=None, yvar_format=var_format)
    put.set_ax_tick_labels(ax=ax, x_rotation=x_rotation, fontsize=fontsize, skip_y_axis=True, **kwargs)
    put.set_ax_xy_labels(ax=ax, xlabel=xlabel, ylabel=ylabel, **kwargs)

    if legend_loc is not None:
        legend_title = None
        if legend_labels is None:
            legend_labels = put.get_legend_lines(data=re_indexed_data,
                                                 legend_stats=legend_stats,
                                                 var_format=var_format)
            if legend_stats in [put.LegendStats.LAST, put.LegendStats.FIRST_LAST_NON_ZERO]:
                legend_title = f"Total: last={var_format.format(re_indexed_data.sum(axis=1).iloc[-1])}"

        put.set_legend(ax=ax,
                       labels=legend_labels,
                       colors=colors,
                       legend_loc=legend_loc,
                       reversed=reversed,
                       ncol=ncol,
                       bbox_to_anchor=bbox_to_anchor,
                       fontsize=fontsize,
                       legend_title=legend_title,
                       **kwargs)

    if title is not None:
        put.set_title(ax=ax, title=title, fontsize=fontsize, **kwargs)

    put.set_spines(ax=ax, **kwargs)

    return fig



"""
factsheet for multi strategy report for cross sectional comparision of strategies
and generating sensetivities to param reports
"""
# packages
import pandas as pd
import matplotlib.pyplot as plt
from typing import Tuple
from enum import Enum

# qis
import qis
import qis.file_utils as fu
import qis.utils.dates as da
import qis.utils.struct_ops as sop
from qis.utils.dates import TimePeriod
from qis.perfstats.config import PerfParams, PerfStat
from qis.perfstats.regime_classifier import BenchmarkReturnsQuantileRegimeSpecs

# portfolio
import qis.portfolio.backtester as bp
from qis.portfolio.multi_portfolio_data import MultiPortfolioData

PERF_PARAMS = PerfParams(freq='W-WED')
REGIME_PARAMS = BenchmarkReturnsQuantileRegimeSpecs(freq='Q')

# use for number years > 5
KWARG_LONG = dict(perf_params=PerfParams(freq='W-WED', freq_reg='Q'),
                  regime_params=BenchmarkReturnsQuantileRegimeSpecs(freq='Q'),
                  x_date_freq='A',
                  date_format='%b-%y')

# use for number years < 3
KWARG_SHORT = dict(perf_params=PerfParams(freq='W-WED', freq_reg='M'),
                   regime_params=BenchmarkReturnsQuantileRegimeSpecs(freq='M'),
                   x_date_freq='Q',
                   date_format='%b-%y')


def generate_multi_portfolio_factsheet(multi_portfolio_data: MultiPortfolioData,
                                       time_period: da.TimePeriod = None,
                                       perf_params: PerfParams = PERF_PARAMS,
                                       regime_params: BenchmarkReturnsQuantileRegimeSpecs = REGIME_PARAMS,
                                       backtest_name: str = None,
                                       figsize: Tuple[float, float] = (8.3, 11.7),  # A4 for portrait
                                       **kwargs
                                       ):
    """
    for portfolio data with structurally different strategies
    """
    regime_benchmark = multi_portfolio_data.benchmark_prices.columns[0]
    benchmark_price = multi_portfolio_data.benchmark_prices[regime_benchmark]

    plot_kwargs = dict(fontsize=5,
                       linewidth=0.5,
                       digits_to_show=1, sharpe_digits=2,
                       weight='normal',
                       markersize=1,
                       framealpha=0.75)
    kwargs = sop.update_kwargs(kwargs, plot_kwargs)

    fig = plt.figure(figsize=figsize, constrained_layout=True)
    gs = fig.add_gridspec(nrows=7, ncols=4, wspace=0.0, hspace=0.0)

    if backtest_name is not None:
        fig.suptitle(backtest_name, fontweight="bold", fontsize=8, color='blue')

    multi_portfolio_data.plot_nav(ax=fig.add_subplot(gs[0, :2]),
                                  time_period=time_period,
                                  benchmark=regime_benchmark,
                                  perf_params=perf_params,
                                  regime_params=regime_params,
                                  title='Cumulative performance',
                                  **kwargs)

    multi_portfolio_data.plot_drawdowns(ax=fig.add_subplot(gs[1, :2]),
                                        time_period=time_period,
                                        benchmark=regime_benchmark,
                                        regime_params=regime_params,
                                        title='Running Drawdowns',
                                        **kwargs)

    multi_portfolio_data.plot_rolling_time_under_water(ax=fig.add_subplot(gs[2, :2]),
                                                       time_period=time_period,
                                                       benchmark=regime_benchmark,
                                                       regime_params=regime_params,
                                                       title='Rolling time under water',
                                                       **kwargs)

    multi_portfolio_data.plot_exposures(ax=fig.add_subplot(gs[3, :2]),
                                        portfolio_idx=0,
                                        time_period=time_period,
                                        benchmark=regime_benchmark,
                                        regime_params=regime_params,
                                        **kwargs)

    multi_portfolio_data.plot_turnover(ax=fig.add_subplot(gs[4, :2]),
                                       time_period=time_period,
                                       benchmark=regime_benchmark,
                                       regime_params=regime_params,
                                       **kwargs)

    multi_portfolio_data.plot_costs(ax=fig.add_subplot(gs[5, :2]),
                                    time_period=time_period,
                                    benchmark=regime_benchmark,
                                    regime_params=regime_params,
                                    **kwargs)

    multi_portfolio_data.plot_factor_betas(ax=fig.add_subplot(gs[6, :2]),
                                           benchmark_prices=multi_portfolio_data.benchmark_prices,
                                           time_period=time_period,
                                           benchmark=regime_benchmark,
                                           regime_params=regime_params,
                                           **kwargs)

    multi_portfolio_data.plot_performance_bars(ax=fig.add_subplot(gs[0, 2]),
                                               perf_params=perf_params,
                                               perf_column=PerfStat.SHARPE,
                                               time_period=time_period,
                                               **sop.update_kwargs(kwargs, dict(fontsize=5)))

    multi_portfolio_data.plot_performance_bars(ax=fig.add_subplot(gs[0, 3]),
                                               perf_params=perf_params,
                                               perf_column=PerfStat.MAX_DD,
                                               time_period=time_period,
                                               **sop.update_kwargs(kwargs, dict(fontsize=5)))

    multi_portfolio_data.plot_periodic_returns(ax=fig.add_subplot(gs[1, 2:]),
                                               heatmap_freq='A',
                                               time_period=time_period,
                                               **sop.update_kwargs(kwargs, dict(date_format='%Y', fontsize=5)))

    multi_portfolio_data.plot_ra_perf_table(ax=fig.add_subplot(gs[2, 2:]),
                                            perf_params=perf_params,
                                            time_period=time_period,
                                            **sop.update_kwargs(kwargs, dict(fontsize=5)))

    multi_portfolio_data.plot_ra_perf_table(ax=fig.add_subplot(gs[3, 2:]),
                                            perf_params=perf_params,
                                            time_period=da.get_time_period_shifted_by_years(time_period=time_period),
                                            **sop.update_kwargs(kwargs, dict(fontsize=5)))

    multi_portfolio_data.plot_regime_data(ax=fig.add_subplot(gs[4, 2:]),
                                          is_grouped=False,
                                          time_period=time_period,
                                          perf_params=perf_params,
                                          regime_params=regime_params,
                                          benchmark=regime_benchmark,
                                          **kwargs)

    multi_portfolio_data.plot_returns_scatter(ax=fig.add_subplot(gs[5, 2:]),
                                              time_period=time_period,
                                              benchmark=regime_benchmark,
                                              **kwargs)

    multi_portfolio_data.plot_corr_table(ax=fig.add_subplot(gs[6, 2:]),
                                         time_period=time_period,
                                         freq='W-WED',
                                         **sop.update_kwargs(kwargs, dict(fontsize=4)))

    return fig


class UnitTests(Enum):
    PERF_ATTRIBUTION = 1
    FACTSHEET = 2


def run_unit_test(unit_test: UnitTests):

    from qis.test_data import load_etf_data
    import qis.models.linear.ra_returns as rar
    prices = load_etf_data()

    time_period = TimePeriod('31Dec2003', '31Dec2022')
    prices = time_period.locate(prices)
    group_data = pd.Series(
        dict(SPY='Equities', QQQ='Equities', EEM='Equities', TLT='Bonds', IEF='Bonds', SHY='Bonds', LQD='Credit',
             HYG='HighYield', GLD='Gold'))

    benchmark_prices = prices[['SPY']]

    spans = [10, 20, 30, 60, 120, 260]
    portfolio_datas = []
    for span in spans:
        ra_returns, weights, ewm_vol = rar.compute_ra_returns(returns=prices.pct_change(), span=span, vol_target=0.15)
        weights = weights.divide(weights.sum(1), axis=0)
        portfolio_data = bp.backtest_model_portfolio(prices=prices,
                                                     weights=time_period.locate(weights),
                                                     is_output_portfolio_data=True,
                                                     ticker=f"VP span-{span}")
        portfolio_data._set_group_data(group_data=group_data, group_order=list(group_data.unique()))
        portfolio_datas.append(portfolio_data)

    multi_portfolio_data = MultiPortfolioData(portfolio_datas,
                                              benchmark_prices=benchmark_prices)

    if unit_test == UnitTests.FACTSHEET:
        fig = generate_multi_portfolio_factsheet(multi_portfolio_data=multi_portfolio_data,
                                                 backtest_name='Vol Parity Portfolio vs Equal Weight',
                                                 time_period=TimePeriod('31Dec2006', '31Dec2022'),
                                                 **KWARG_LONG)

        fu.save_figs_to_pdf(figs=[fig],
                            file_name=f"multistrategy_factsheet",
                            orientation='landscape',
                            local_path=qis.local_path.get_output_path())

    plt.show()


if __name__ == '__main__':

    unit_test = UnitTests.FACTSHEET

    is_run_all_tests = False
    if is_run_all_tests:
        for unit_test in UnitTests:
            run_unit_test(unit_test=unit_test)
    else:
        run_unit_test(unit_test=unit_test)
